////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2018 The University of North Carolina at Chapel Hill.
// All Rights Reserved.
//
// Permission to use, copy, modify and distribute this software and its
// documentation for educational, research and non-profit purposes, without
// fee, and without a written agreement is hereby granted, provided that the
// above copyright notice and the following three paragraphs appear in all
// copies.
//
// IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE
// LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
// USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY
// OF NORTH CAROLINA HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGES.
//
// THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIM ANY
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
// PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
// NORTH CAROLINA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
// UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The authors may be contacted via:
//
// US Mail: Real-Time Systems Group at UNC
// Department of Computer Science
// Sitterson Hall
// University of N. Carolina
// Chapel Hill, NC 27599-3175
//
// EMail: nemitz@cs.unc.edu; tamert@cs.unc.edu; anderson@cs.unc.edu
//
////////////////////////////////////////////////////////////////////////////////

#ifndef __Q_LOCK_H__
#define __Q_LOCK_H__

#include <assert.h>

#include "mem.h"
#include "cycles.h"

typedef struct __mcs_node
{
	volatile int	            blocked;
	volatile struct __mcs_node*  next;	
} mcs_node_t;

typedef struct
{
	volatile mcs_node_t*	tail;
} mcs_lock_t;

static inline void mcs_init(mcs_lock_t* lock)
{
    lock->tail = NULL;
}

static inline uint64_t *mcs_lock(mcs_lock_t* lock, mcs_node_t *self)
{
    mcs_node_t* prev;

    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0;

    start = get_cycles();

    self->next = NULL;
    self->blocked = 1;

    do
    {
        prev = (mcs_node_t*) (lock->tail);
    } while (!__sync_bool_compare_and_swap(&lock->tail, prev, self));

    if (prev)
    {
        prev->next = self;

        end = get_cycles();
        ret[0] += end - start;
        start = get_cycles();

        while (self->blocked)
        {
            cpu_relax();
            __sync_synchronize();
        }

        end = get_cycles();
        ret[1] += end - start;
    }
    else
    {
        end = get_cycles();
        ret[0] += end - start;
    }

    return ret;
}

static inline uint64_t *mcs_unlock(mcs_lock_t* lock, mcs_node_t *self)
{
    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0;

    start = get_cycles();

    if (!self->next && !__sync_bool_compare_and_swap(&lock->tail, self, 0))
    {
        while (!self->next)
        {
            cpu_relax();
            __sync_synchronize();
        }
    }

    if (self->next)
    {
        self->next->blocked = 0;
    }

    end = get_cycles();
    ret[0] += end - start;

    return ret;
}

#endif
