////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2018 The University of North Carolina at Chapel Hill.
// All Rights Reserved.
//
// Permission to use, copy, modify and distribute this software and its
// documentation for educational, research and non-profit purposes, without
// fee, and without a written agreement is hereby granted, provided that the
// above copyright notice and the following three paragraphs appear in all
// copies.
//
// IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE
// LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
// USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY
// OF NORTH CAROLINA HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGES.
//
// THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIM ANY
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
// PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
// NORTH CAROLINA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
// UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The authors may be contacted via:
//
// US Mail: Real-Time Systems Group at UNC
// Department of Computer Science
// Sitterson Hall
// University of N. Carolina
// Chapel Hill, NC 27599-3175
//
// EMail: nemitz@cs.unc.edu; tamert@cs.unc.edu; anderson@cs.unc.edu
//
////////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>

#include "cycles.h"
#include "crnlp_general_flls.h"
#include "nesting.h"
#include "mem.h"
#include "queue.h"

void crnlp_general_flls_init(crnlp_general_flls_lock_t* lock)
{
    int i, j;

    for (i = 0; i < NUM_SERVERS; i++)
    {
        // Initialize resource queues
        for (j = 0; j < RES_COUNT; j++)
        {
            lock->servers[i].queues[j].head = NULL;
            lock->servers[i].queues[j].tail = NULL;
        }

        // Initialize the requests
        for (j = 0; j < QUEUE_LENGTH; j++)
        {
            lock->servers[i].requests[j].requested = 0;
            lock->servers[i].requests[j].length = 0;
            lock->servers[i].requests[j].request_id = j;
            lock->servers[i].requests[j].enabled = 0;
            lock->servers[i].requests[j].is_head = 0;
            lock->servers[i].requests[j].needs_service = 0;
        }

        lock->servers[i].pending_requests = 0;
        lock->servers[i].satisfied_requests = 0;
        lock->servers[i].enabled = 0;

        // Initialize the RR lock logic components
        lock->servers[i].rr_wave = 0;

        lock->servers[i].service_index = 0;
        lock->servers[i].server_exists = 0;
        lock->servers[i].other_server_bits = 0;
    }

    // Initialize the RR lock logic components
    lock->S = 0;
}

uint64_t *crnlp_general_flls_lock(crnlp_general_flls_lock_t* lock, uint8_t ind)
{
    crnlp_general_flls_server_lock_t *server_lock;
    crnlp_general_flls_request_t *self, *other;
    volatile crnlp_general_flls_queue_t *queue, *queue2;
    volatile struct crnlp_general_flls_node_t *node, *next_node;

    int r, s;
    uint8_t start_in_sched, i, j;
    uint8_t is_lock_server, server_id;
    unsigned long one, tmp, tmp2;
    unsigned int other_server_bits;

    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0;

    start = get_cycles();

    one = 1;

    server_id = ind / (QUEUE_LENGTH / NUM_SERVERS);
    server_lock = &(lock->servers[server_id]);
    self = &(server_lock->requests[ind]);

    self->enabled = 0;
    self->is_head = 0;

    // Mark this request as needing to be serviced for a lock
    self->done = 0;
    mb();
    self->needs_service = 1;

    // Check whether a lock server exists (and potentially become it)
    is_lock_server = 0;
    if (!server_lock->server_exists)
    {
        is_lock_server = __sync_bool_compare_and_swap(&(server_lock->server_exists), 0, 1);
    }

    // Wait until this request is serviced, or this core becomes the lock server
    // (if it is already the lock server, skip this; it will service itself later)
    while (!is_lock_server && self->needs_service != 0)
    {
        // Try to become the lock server (in the case the previous lock server
        // completed without servicing this request)
        if (!server_lock->server_exists)
        {
            is_lock_server = __sync_bool_compare_and_swap(&(server_lock->server_exists), 0, 1);
        }
    }

    // If this core is not the lock server, then we switch to measuring
    // blocking time (the request must have been serviced)
    if (!is_lock_server)
    {
        end = get_cycles();
        ret[0] += end - start;
        start = get_cycles();
    }

    // Wait until this request is satisfied, or this core becomes the lock server
    // (if it is already the lock server, skip this; it will eventually become
    // satisfied, after potentially servicing other requests)
    while (!is_lock_server && self->enabled == 0) 
    {
        // Try to become the lock server (in order to help service requests
        // while waiting to be satisfied)
        if (!server_lock->server_exists)
        {
            is_lock_server = __sync_bool_compare_and_swap(&(server_lock->server_exists), 0, 1);
        }
    }

    // If this core is not the lock server, then the request must be satisfied
    if (!is_lock_server)
    {
        end = get_cycles();
        ret[1] += end - start;

        return ret;
    }

    // Otherwise, help out as long as this request has not yet been serviced
    // and is not yet satisfied
    while (self->needs_service != 0 || self->enabled == 0)
    {
        // Check if the next request needs to be serviced
        i = server_lock->service_index;
        if (server_lock->requests[i].needs_service == 1)
        {
            // Service request i
            other = &(server_lock->requests[i]);
            if (other->done == 0)
            {
                // Lock code
                uint64_t start_in_sched, prev_start, potential_start;

                // Find the earliest possible start time
                start_in_sched = get_cycles();
                prev_start = start_in_sched;

                while (1)
                {
                    // Loop through the required resources
                    tmp = other->requested;
                    r = ffsl(tmp);
                    while (r != 0)
                    {
                        r = r - 1;

                        // For this resource, find the earliest possible start time after
                        // the current start_in_sched within the resource queue
                        queue = &(server_lock->queues[r]);

                        if (queue->head)
                        {
                            // If this request cannot be satisfied before the first node,
                            // we need to find the node it will follow
                            if ((queue->head->request->enabled == 1) ||
                                (get_cycles() + other->length >= queue->head->start_time))
                            {
                                node = queue->head->next;

                                while (node)
                                {
                                    potential_start = node->prev->start_time +
                                                      node->prev->request->length;
                                    potential_start = potential_start < start_in_sched ?
                                                          start_in_sched :
                                                          potential_start;
                                    if (potential_start + other->length < node->start_time)
                                    {
                                        // This location is delay-preserving!
                                        start_in_sched = potential_start;
                                        break;
                                    }
                                    else
                                    {
                                        // We need to continue looking in the schedule
                                        node = node->next;
                                    }
                                }

                                // If node is NULL, we belong at the tail of the queue
                                if (!node)
                                {
                                    potential_start = queue->tail->start_time +
                                                      queue->tail->request->length;
                                    if (potential_start > start_in_sched)
                                    {
                                        start_in_sched = potential_start;
                                    }
                                }
                            }
                        }

                        tmp &= ~(one << r);
                        r = ffsl(tmp);
                    }

                    if (prev_start == start_in_sched)
                    {
                        // If the selected time did not change, we are done looping
                        break;
                    }
                    else
                    {
                        // Otherwise, update prev_start and go around again
                        prev_start = start_in_sched;
                    }
                }

                // Update pointers to insert the request into the queue for each
                // required resource
                j = 0;
                int is_head = 1;
                tmp = other->requested;
                r = ffsl(tmp);
                while (r != 0)
                {
                    r = r - 1;

                    // Grab a queue node from the pool
                    node = &(server_lock->nodes[i][j++]);

                    node->request = other;
                    node->start_time = start_in_sched;

                    node->prev = NULL;
                    node->next = NULL;

                    // Insert this node into the queue
                    queue = &(server_lock->queues[r]);
                    if (!queue->head)
                    {
                        queue->head = node;
                        queue->tail = node;
                    }
                    else
                    {
                        // Find the node it will precede
                        next_node = queue->head;
                        while (next_node && (next_node->start_time <= start_in_sched))
                        {
                            next_node = next_node->next;
                        }

                        // Insert the new node before next_node
                        if (!next_node)
                        {
                            queue->tail->next = node;
                            node->prev = queue->tail;

                            queue->tail = node;

                            is_head = 0;
                        }
                        else if (next_node == queue->head)
                        {
                            queue->head = node;

                            next_node->prev = node;
                            node->next = next_node;
                        }
                        else
                        {
                            next_node->prev->next = node;
                            node->prev = next_node->prev;

                            next_node->prev = node;
                            node->next = next_node;

                            is_head = 0;
                        }
                    }

                    tmp &= ~(one << r);
                    r = ffsl(tmp);
                }

                // Update the count of pending requests
                server_lock->pending_requests += 1;

                mb();

                // If the server was idle, update the phase and mark the
                // server present
                if (server_lock->pending_requests == 1)
                {
                    server_lock->rr_wave = ~(server_lock->rr_wave);
                    if (server_id == 0)
                    {
                        other_server_bits = __sync_fetch_and_add(
                                                &lock->S,
                                                APRES | (server_lock->rr_wave & APHID));
                        server_lock->other_server_bits = other_server_bits & BBITS;
                    }
                    else
                    {
                        other_server_bits = __sync_fetch_and_add(
                                                &lock->S,
                                                BPRES | (server_lock->rr_wave & BPHID));
                        server_lock->other_server_bits = other_server_bits & ABITS;
                    }

                    mb();
                }

                // Mark this request as serviced
                other->needs_service = 0;
            }
            else
            {
                // Unlock code

                // Iterate through the resource queues for this request
                tmp = other->requested;
                r = ffsl(tmp);
                while (r != 0)
                {
                    r = r - 1;

                    queue = &(server_lock->queues[r]);
                    node = queue->head;

                    // Update the head
                    queue->head = queue->head->next;

                    // Update the new head's prev pointer
                    if (queue->head)
                    {
                        queue->head->prev = NULL;
                    }
                    else
                    {
                        queue->tail = NULL;
                    }

                    tmp &= ~(one << r);
                    r = ffsl(tmp);
                }

                // Decrement the pending and satisfied request counts,
                // as this request is now finished
                server_lock->pending_requests -= 1;
                server_lock->satisfied_requests -= 1;

                mb();

                // Mark this request as serviced
                other->needs_service = 0;

                if (server_lock->satisfied_requests == 0)
                {
                    server_lock->enabled = 0;

                    // Clear the XBITS for this server, allowing another
                    // server to become enabled
                    if (server_id == 0)
                    {
                        __sync_fetch_and_and(&lock->S, ~ABITS);
                    }
                    else
                    {
                        __sync_fetch_and_and(&lock->S, ~BBITS);
                    }

                    // If there are still requests on this server, update
                    // the wave and present bits
                    if (server_lock->pending_requests > 0)
                    {
                        server_lock->rr_wave = ~(server_lock->rr_wave);
                        if (server_id == 0)
                        {
                            other_server_bits = __sync_fetch_and_add(
                                                    &lock->S,
                                                    APRES | (server_lock->rr_wave & APHID));
                            server_lock->other_server_bits = other_server_bits & BBITS;
                        }
                        else
                        {
                            other_server_bits = __sync_fetch_and_add(
                                                    &lock->S,
                                                    BPRES | (server_lock->rr_wave & BPHID));
                            server_lock->other_server_bits = other_server_bits & ABITS;
                        }

                        mb();
                    }
                }
            }

            // If this request was just serviced, then we switch to measuring
            // blocking time
            if (i == ind)
            {
                end = get_cycles();
                ret[0] += end - start;
                start = get_cycles();
            }
        }

        // If is server is not enabled, check the RR logic to determine
        // if it should be enabled (make sure the last phase is done
        // being fully processed, i.e. satisfied_requests is back to 0)
        if ((server_lock->pending_requests > 0) &&
            (server_lock->enabled != 1) &&
            (server_lock->satisfied_requests == 0))
        {
            uint8_t should_enable = 0;
            if ((server_id == 0) &&
                ((server_lock->other_server_bits == 0) ||
                 (server_lock->other_server_bits != (lock->S & BBITS))))
            {
                should_enable = 1;
            }
            else if ((server_id == 1) &&
                ((server_lock->other_server_bits == 0) ||
                 (server_lock->other_server_bits != (lock->S & ABITS))))
            {
                should_enable = 1;
            }

            if (should_enable == 1)
            {
                // Mark the server as enabled
                server_lock->enabled = 1;

                uint8_t to_enable[QUEUE_LENGTH];
                memset(to_enable, 0, QUEUE_LENGTH * sizeof(uint8_t));
                unsigned my_idx = 0;

                // Check the heads of the resource queues to determine
                // if they are now enabled
                for (r = 0; r < RES_COUNT; r++)
                {
                    queue = &(server_lock->queues[r]);

                    // If the queue is empty or the request at the head
                    // is enabled, skip it
                    //if (queue->head && queue->head->request->enabled == 0)
                    if (queue->head && queue->head->request->is_head == 0)
                    {
                        // Check if the head of this queue is now enabled,
                        // by checking the queues for this request
                        int is_enabled = 1;

                        tmp2 = queue->head->request->requested;
                        s = ffsl(tmp2);
                        while (s != 0)
                        {
                            s = s - 1;

                            queue2 = &(server_lock->queues[s]);

                            // If this request is not the head, it can't
                            // be enabled
                            if (queue->head->request != queue2->head->request)
                            {
                                is_enabled = 0;
                                break;
                            }

                            tmp2 &= ~(one << s);
                            s = ffsl(tmp2);
                        }

                        // If the request at the head is enabled, set
                        // its bit
                        if (is_enabled == 1)
                        {
                            to_enable[my_idx++] = queue->head->request->request_id;
                            queue->head->request->is_head = 1;
                        }
                    }
                }

                // Now do the enabling
                server_lock->satisfied_requests = my_idx;
                for (r = 0; r < my_idx; r++)
                {
                    server_lock->requests[to_enable[r]].enabled = 1;
                }
            }
        }

        // Move onto the next request
        server_lock->service_index = (i + 1) % QUEUE_LENGTH;
    }

    // Set the server to not exist
    server_lock->server_exists = 0;

    end = get_cycles();
    ret[1] += end - start;

    // Satisfied!
    return ret;
}

uint64_t *crnlp_general_flls_unlock(crnlp_general_flls_lock_t* lock, uint8_t ind)
{
    crnlp_general_flls_server_lock_t *server_lock;
    crnlp_general_flls_request_t *self, *other;
    volatile crnlp_general_flls_queue_t *queue, *queue2;
    volatile struct crnlp_general_flls_node_t *node, *next_node;
    uint8_t is_lock_server, server_id;

    int r, s;
    unsigned long one, tmp, tmp2;
    unsigned int other_server_bits;

    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0;

    start = get_cycles();

    one = 1;

    server_id = ind / (QUEUE_LENGTH / NUM_SERVERS);
    server_lock = &(lock->servers[server_id]);
    self = &(server_lock->requests[ind]);

    // Mark this request as needing to be serviced for an unlock
    self->done = 1;
    mb();
    self->needs_service = 1;

    // Check whether a lock server exists (and potentially become it)
    is_lock_server = 0;
    if (!server_lock->server_exists)
    {
        is_lock_server = __sync_bool_compare_and_swap(&(server_lock->server_exists), 0, 1);
    }

    // Wait until this request is serviced, or this core becomes the lock server
    // (if it is already the lock server, skip this; it will service itself later)
    while (!is_lock_server && self->needs_service != 0)
    {
        // Try to become the lock server (in the case the previous lock server
        // completed without servicing this request)
        if (!server_lock->server_exists)
        {
            is_lock_server = __sync_bool_compare_and_swap(&(server_lock->server_exists), 0, 1);
        }
    }

    // If this core is the lock server and has not been serviced, be
    // greedy and just service this unlock
    if (is_lock_server && self->needs_service == 0)
    {
        // Oops, this core shouldn't be the server
        server_lock->server_exists = 0;
    }
    else if (is_lock_server)
    {
        // Iterate through the resource queues for this request
        tmp = self->requested;
        r = ffsl(tmp);
        while (r != 0)
        {
            r = r - 1;

            queue = &(server_lock->queues[r]);
            node = queue->head;

            // Update the head
            queue->head = queue->head->next;

            // Update the new head's prev pointer
            if (queue->head)
            {
                queue->head->prev = NULL;
            }
            else
            {
                queue->tail = NULL;
            }

            tmp &= ~(one << r);
            r = ffsl(tmp);
        }

        // Decrement the pending and satisfied request counts,
        // as this request is now finished
        server_lock->pending_requests -= 1;
        server_lock->satisfied_requests -= 1;

        mb();

        // Mark this request as serviced
        self->needs_service = 0;

        if (server_lock->satisfied_requests == 0)
        {
            server_lock->enabled = 0;

            // Clear the XBITS for this server, allowing another
            // server to become enabled
            if (server_id == 0)
            {
                __sync_fetch_and_and(&lock->S, ~ABITS);
            }
            else
            {
                __sync_fetch_and_and(&lock->S, ~BBITS);
            }

            // If there are still requests on this server, update
            // the wave and present bits
            if (server_lock->pending_requests > 0)
            {
                server_lock->rr_wave = ~(server_lock->rr_wave);
                if (server_id == 0)
                {
                    other_server_bits = __sync_fetch_and_add(
                                            &lock->S,
                                            APRES | (server_lock->rr_wave & APHID));
                    server_lock->other_server_bits = other_server_bits & BBITS;
                }
                else
                {
                    other_server_bits = __sync_fetch_and_add(
                                            &lock->S,
                                            BPRES | (server_lock->rr_wave & BPHID));
                    server_lock->other_server_bits = other_server_bits & ABITS;
                }

                mb();
            }
        }

        // Set the server to not exist
        server_lock->server_exists = 0;
    }

    end = get_cycles();
    ret[0] += end - start;

    return ret;
}

